# -*- coding: utf-8 -*-
# Copyright 2020 The PsiZ Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for loading internally pre-defined datasets.

Functions:
    load: Load observations for the requested dataset.

Notes:
    The dataset will only be downloaded from the server if it does not
    exist locally. If it it already exists locally, no download will
    take place. If you would like to force a download, delete the
    existing local copy.

"""

import collections
import copy
import os
from pathlib import Path
import shutil
import sys
import tarfile
import time
from urllib import request
import zipfile

import h5py
import numpy as np
import pandas as pd
import six
from six.moves.urllib.error import HTTPError
from six.moves.urllib.error import URLError
from six.moves.urllib.request import urlopen
from six.moves.urllib.request import urlretrieve

import psiz.catalog
import psiz.trials

HOST_URL = "https://www.psiz.org/datasets/"


def load(
        dataset_name, cache_subdir='datasets', cache_dir=None,
        verbose=0):
    """Load observations and catalog for the requested hosted dataset.

    Arguments:
        dataset_name: The name of the hosted dataset.
        cache_subdir (optional): The subdirectory where downloaded
            datasets are cached.
        cache_dir (optional): The cache directory for PsiZ.
        verbose (optional): Controls the verbosity of printed dataset summary.

    Returns:
        obs: An Observations object.
        catalog: A catalog object containing information regarding the
            stimuli used to collect observations.

    """
    # Load from download cache.
    if cache_dir is None:
        cache_dir = Path.home() / Path('.psiz')
    else:
        cache_dir = Path(cache_dir)
    dataset_path = cache_dir / Path(cache_subdir, dataset_name)
    if not dataset_path.exists():
        dataset_path.mkdir(parents=True)

    obs = _fetch_obs(dataset_name, cache_dir, cache_subdir)
    catalog = _fetch_catalog(dataset_name, cache_dir, cache_subdir)

    if verbose > 0:
        print("Dataset Summary")
        print('  n_stimuli: {0}'.format(catalog.n_stimuli))
        print('  n_trial: {0}'.format(obs.n_trial))

    return (obs, catalog)


def _fetch_catalog(dataset_name, cache_dir, cache_subdir):
    """Fetch catalog for the requested dataset.

    Arguments:
        dataset_name: The name of the dataset to load.
        cache_dir: The cache directory for PsiZ.
        cache_subdir: The subdirectory where downloaded datasets are
            cached.

    Returns:
        catalog: A Catalog object.

    """
    fname = "catalog.hdf5"

    dataset_exists = True
    if dataset_name == "birds-12":
        origin = HOST_URL + "birds-12/" + fname
    elif dataset_name == "birds-16":
        origin = HOST_URL + "birds-16/" + fname
    elif dataset_name == "lesions":
        origin = HOST_URL + "lesions/" + fname
    elif dataset_name == "rocks_Nosofsky_etal_2016":
        origin = HOST_URL + "rocks_Nosofsky_etal_2016/" + fname
    else:
        dataset_exists = False

    if dataset_exists:
        path = _get_file(
            os.path.join(dataset_name, fname), origin,
            cache_subdir=cache_subdir, extract=True,
            cache_dir=cache_dir
        )
        catalog = psiz.catalog.load(path)
    else:
        raise ValueError(
            'The requested dataset `{0}` may not exist since the '
            'corresponding catalog.hdf5 file does not '
            'exist.'.format(dataset_name)
        )

    return catalog


def _fetch_obs(dataset_name, cache_dir, cache_subdir):
    """Fetch observations for the requested dataset.

    Arguments:
        dataset_name: The name of the dataset to load.
        cache_dir: The cache directory for PsiZ.
        cache_subdir: The subdirectory where downloaded datasets are
            cached.

    Returns:
        obs: An Observations object.

    """
    fname = 'obs.hdf5'

    dataset_exists = True
    if dataset_name == "birds-12":
        origin = HOST_URL + "birds-12/" + fname
    elif dataset_name == "birds-16":
        origin = HOST_URL + "birds-16/" + fname
    elif dataset_name == "lesions":
        origin = HOST_URL + "lesions/" + fname
    elif dataset_name == "rocks_Nosofsky_etal_2016":
        origin = HOST_URL + "rocks_Nosofsky_etal_2016/" + fname
    else:
        dataset_exists = False

    if dataset_exists:
        path = _get_file(
            os.path.join(dataset_name, fname), origin,
            cache_subdir=cache_subdir, extract=True, cache_dir=cache_dir
        )
        obs = psiz.trials.load(path)
    else:
        raise ValueError(
            'The requested dataset `{0}` may not exist since the '
            'corresponding obs.hdf5 file does not '
            'exist.'.format(dataset_name)
        )

    return obs


def _extract_archive(file_path, path='.', archive_format='auto'):
    """Extract an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    Arguments:
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    Returns:
        True if a match was found and an archive extraction was completed,
        False otherwise.

    """
    if archive_format is None:
        return False
    if archive_format is 'auto':
        archive_format = ['tar', 'zip']
    if isinstance(archive_format, six.string_types):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type is 'tar':
            open_fn = tarfile.open
            is_match_fn = tarfile.is_tarfile
        if archive_type is 'zip':
            open_fn = zipfile.ZipFile
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError,
                        KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def _get_file(
        fname, origin, untar=False, cache_subdir='datasets', extract=False,
        archive_format='auto', cache_dir=None):
    """Download a file from a URL if it not already in the cache."""
    if cache_dir is None:
        cache_dir = os.path.join(os.path.expanduser('~'), '.psiz')

    datadir_base = os.path.expanduser(cache_dir)
    if not os.access(datadir_base, os.W_OK):
        datadir_base = os.path.join('/tmp', '.psiz')
    datadir = os.path.join(datadir_base, cache_subdir)
    if not os.path.exists(datadir):
        os.makedirs(datadir)

    if untar:
        untar_fpath = os.path.join(datadir, fname)
        fpath = untar_fpath + '.tar.gz'
    else:
        fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        download = False
    else:
        download = True

    if download:
        print('Downloading data from', origin)

        class ProgressTracker(object):
            # Maintain progbar for the lifetime of download.
            # This design was chosen for Python 2.7 compatibility.
            progbar = None

        def dl_progress(count, block_size, total_size):
            if ProgressTracker.progbar is None:
                if total_size is -1:
                    total_size = None
                ProgressTracker.progbar = Progbar(total_size)
            else:
                ProgressTracker.progbar.update(count * block_size)

        error_msg = 'URL fetch failure on {}: {} -- {}'
        try:
            try:
                urlretrieve(origin, fpath, dl_progress)
            except HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise

    if untar:
        if not os.path.exists(untar_fpath):
            _extract_archive(fpath, datadir, archive_format='tar')
        return untar_fpath

    if extract:
        _extract_archive(fpath, datadir, archive_format)

    return fpath


class Progbar(object):
    """Displays a progress bar.

    Arguments:
        target: Total number of steps expected, None if unknown.
        width: Progress bar width on screen.
        verbose: Verbosity mode, 0 (silent), 1 (verbose), 2 (semi-verbose)
        stateful_metrics: Iterable of string names of metrics that
            should *not* be averaged over time. Metrics in this list
            will be displayed as-is. All others will be averaged
            by the progbar before display.
        interval: Minimum visual progress update interval (in seconds).

    """

    def __init__(
            self, target, width=30, verbose=1, interval=0.05,
            stateful_metrics=None):
        """Initialize."""
        self.target = target
        self.width = width
        self.verbose = verbose
        self.interval = interval
        if stateful_metrics:
            self.stateful_metrics = set(stateful_metrics)
        else:
            self.stateful_metrics = set()

        self._dynamic_display = ((hasattr(sys.stdout, 'isatty') and
                                  sys.stdout.isatty()) or
                                 'ipykernel' in sys.modules)
        self._total_width = 0
        self._seen_so_far = 0
        self._values = collections.OrderedDict()
        self._start = time.time()
        self._last_update = 0

    def update(self, current, values=None):
        """Update the progress bar.

        Arguments
            current: Index of current step.
            values: List of tuples:
                `(name, value_for_last_step)`.
                If `name` is in `stateful_metrics`,
                `value_for_last_step` will be displayed as-is.
                Else, an average of the metric over time will be displayed.

        """
        values = values or []
        for k, v in values:
            if k not in self.stateful_metrics:
                if k not in self._values:
                    self._values[k] = [v * (current - self._seen_so_far),
                                       current - self._seen_so_far]
                else:
                    self._values[k][0] += v * (current - self._seen_so_far)
                    self._values[k][1] += (current - self._seen_so_far)
            else:
                # Stateful metrics output a numeric value.  This representation
                # means "take an average from a single value" but keeps the
                # numeric formatting.
                self._values[k] = [v, 1]
        self._seen_so_far = current

        now = time.time()
        info = ' - %.0fs' % (now - self._start)
        if self.verbose == 1:
            if (now - self._last_update < self.interval and
                    self.target is not None and current < self.target):
                return

            prev_total_width = self._total_width
            if self._dynamic_display:
                sys.stdout.write('\b' * prev_total_width)
                sys.stdout.write('\r')
            else:
                sys.stdout.write('\n')

            if self.target is not None:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = '%%%dd/%d [' % (numdigits, self.target)
                bar = barstr % current
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
            else:
                bar = '%7d/Unknown' % current

            self._total_width = len(bar)
            sys.stdout.write(bar)

            if current:
                time_per_unit = (now - self._start) / current
            else:
                time_per_unit = 0
            if self.target is not None and current < self.target:
                eta = time_per_unit * (self.target - current)
                if eta > 3600:
                    eta_format = ('%d:%02d:%02d' %
                                  (eta // 3600, (eta % 3600) // 60, eta % 60))
                elif eta > 60:
                    eta_format = '%d:%02d' % (eta // 60, eta % 60)
                else:
                    eta_format = '%ds' % eta

                info = ' - ETA: %s' % eta_format
            else:
                if time_per_unit >= 1:
                    info += ' %.0fs/step' % time_per_unit
                elif time_per_unit >= 1e-3:
                    info += ' %.0fms/step' % (time_per_unit * 1e3)
                else:
                    info += ' %.0fus/step' % (time_per_unit * 1e6)

            for k in self._values:
                info += ' - %s:' % k
                if isinstance(self._values[k], list):
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if abs(avg) > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                else:
                    info += ' %s' % self._values[k]

            self._total_width += len(info)
            if prev_total_width > self._total_width:
                info += (' ' * (prev_total_width - self._total_width))

            if self.target is not None and current >= self.target:
                info += '\n'

            sys.stdout.write(info)
            sys.stdout.flush()

        elif self.verbose == 2:
            if self.target is None or current >= self.target:
                for k in self._values:
                    info += ' - %s:' % k
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if avg > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                info += '\n'

                sys.stdout.write(info)
                sys.stdout.flush()

        self._last_update = now

    def add(self, n, values=None):
        """Add progress."""
        self.update(self._seen_so_far + n, values)




# -*- coding: utf-8 -*-
# Copyright 2020 The PsiZ Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module of utility functions.

Classes:
    ProgressBar: A progress bar displayed in the terminal.

Functions:
    pairwise_matrix: Return the similarity matrix characterizing
        the embedding.
    similarity_matrix: Return the similarity matrix characterizing
        the embedding. Deprecated use pairwise_matrix.
    matrix_comparison: Compute correlation between two matrices.
    compare_models: Compare the similarity structure between two
        embedding models.
    rotation_matrix: Returns a two-dimensional rotation matrix.
    affine_transformation: Performs an affine transformation on a set
        of points.
    procrustean_solution: Attempt to allign two embeddings by finding
        a Procrustean solution.
"""

import copy
import datetime
import time

import numpy as np
import pandas as pd
from scipy.optimize import minimize
from scipy.stats import pearsonr
from sklearn.metrics import r2_score


def assess_convergence(
        obs, model, n_stimuli, n_dim, n_partition=10, n_back=2, n_shuffle=3,
        n_restart=50, score='r2', verbose=0):
    """Evaluate if a sufficient number of observations have been collected.

    In general, more observations improve model inference. However,
    there are generally diminishing returns to collecting more data.
    This function determines the impact of adding a little bit more
    data. Ideally, you should have sufficient data that adding more
    does not dramatically alter the structure of the inferred
    embedding. Structural changes are assessed by comparing the pair-
    wise similarity matrix of different embeddings.

    Arguments:
        obs: An Observations object.
        model: A PsychologicalEmbedding object.
        n_stimuli: The number of stimuli.
        n_dim: The dimensionality of the embedding.
        n_partition (optional): The number of partitions.
        n_back (optional):  The number of partitions to evaluate. Can
            be [1, n_partition-1].
        n_shuffle (optional): The number of times to shuffle and
            repeat tha analysis.
        n_restart (optional): The number of restarts to use when
            fitting the embeddings.
        score (optional): Measure to use when comparing two similarity
            matrices.
        verbose (optional): Verbosity.

    Returns:
        summary: Dictionary of results.
            n_trial_array: Number of trials in each set.

    """
    # Check arguments.
    n_back = np.maximum(n_back, 1)
    n_back = np.minimum(n_back, n_partition-1)

    n_trial_array = np.linspace(0, obs.n_trial, n_partition + 1)
    n_trial_array = np.ceil(n_trial_array[1:]).astype(int)

    val = np.ones([n_shuffle, n_partition]) * np.nan
    for i_shuffle in range(n_shuffle):
        if verbose > 0:
            print('  Shuffle {0}'.format(i_shuffle + 1))
        # Randomize observations.
        rand_idx = np.random.permutation(obs.n_trial)
        obs = obs.subset(rand_idx)

        # Infer embeddings for an increasing number of observations.
        first_part = True
        emb_list = [None] * n_partition
        for i_part in range(n_partition-n_back-1, n_partition):
            include_idx = np.arange(0, n_trial_array[i_part])
            curr_obs = obs.subset(include_idx)
            curr_emb = model(n_stimuli, n_dim=n_dim)
            curr_emb.fit(curr_obs, n_restart=n_restart, verbose=verbose)
            emb_list[i_part] = curr_emb

            if not first_part:
                simmat_0 = similarity_matrix(
                    emb_list[i_part - 1].similarity,
                    emb_list[i_part - 1].z
                )
                simmat_1 = similarity_matrix(
                    emb_list[i_part].similarity, emb_list[i_part].z
                )
                val[i_shuffle, i_part] = matrix_comparison(
                    simmat_0, simmat_1, score=score
                )
                if verbose > 0:
                    print('    {0} | measure: {1:.2f}'.format(
                        i_part, val[i_shuffle, i_part]
                    ))
            else:
                first_part = False

    return {"n_trial_array": n_trial_array, "val": val, "measure": score}


def similarity_matrix(similarity_fn, z):
    """Return a pairwise similarity matrix.

    Returns:
        A 2D array where element s_{i,j} indicates the similarity
            between the ith and jth stimulus.

    """
    n_stimuli = z.shape[0]

    xg = np.arange(n_stimuli)
    a, b = np.meshgrid(xg, xg)
    a = a.flatten()
    b = b.flatten()

    z_a = z[a, :]
    z_b = z[b, :]
    simmat = similarity_fn(z_a, z_b)
    simmat = simmat.reshape(n_stimuli, n_stimuli)
    return simmat


pairwise_matrix = similarity_matrix


def matrix_comparison(mat_a, mat_b, score='r2', elements='upper'):
    """Return a comparison score between two square matrices.

    Arguments:
        mat_a: A square matrix.
        mat_b: A square matrix the same size as mat_a
        score (optional): The type of comparison to use.
        elements (optional): Which elements to use in the computation.
            The options are upper triangular elements (upper), lower
            triangular elements (lower), or off-diagonal elements
            (off).

    Returns:
        The comparison score.

    Notes:
        'r2'
        When computing R^2 values of two similarity matrices, it is
        assumed, by definition, that the corresponding diagonal
        elements are the same between the two matrices being compared.
        Therefore, the diagonal elements are not included in the R^2
        computation to prevent inflating the R^2 value. On a similar
        note, including both the upper and lower triangular portion
        does not artificially inflate R^2 values for symmetric
        matrices.

    """
    n_row = mat_a.shape[0]
    idx_upper = np.triu_indices(n_row, 1)
    idx_lower = np.triu_indices(n_row, 1)
    if elements == 'upper':
        idx = idx_upper
    elif elements == 'lower':
        idx = idx_lower
    elif elements == 'off':
        idx = (
            np.hstack((idx_upper[0], idx_lower[0])),
            np.hstack((idx_upper[1], idx_lower[1])),
        )
    else:
        raise ValueError(
            'The argument to `elements` must be "upper", "lower", or "off".')

    if score == 'pearson':
        score, _ = pearsonr(mat_a[idx], mat_b[idx])
    elif score == 'r2':
        rho, _ = pearsonr(mat_a[idx], mat_b[idx])
        score = rho**2
        # score = r2_score(mat_a[idx], mat_b[idx])
    elif score == 'mse':
        score = np.mean((mat_a[idx] - mat_b[idx])**2)
    else:
        raise ValueError(
            'The provided `score` argument is not valid.')
    return score


def compare_models(
        model_a, model_b, group_id_a=0, group_id_b=0, score='r2',
        elements='upper'):
    """Compare two psychological embeddings.

    Arguments:
        model_a:  A psychological embedding model.
        model_b:  A psychological embedding model.
        group_id_a (optional):  A particular group ID to use when
            computing the similarity matrix for model_a.
        group_id_b (optional):  A particular group ID to use when
            computing the similarity matrix for model_b.

    Returns:
        R^2 value of the two similarity matrices derived from each
            embedding.

    """
    def sim_func_a(z_q, z_ref):
        return model_a.similarity(z_q, z_ref, group_id=group_id_a)

    def sim_func_b(z_q, z_ref):
        return model_b.similarity(z_q, z_ref, group_id=group_id_b)

    simmat_a = similarity_matrix(sim_func_a, model_a.z)
    simmat_b = similarity_matrix(sim_func_b, model_b.z)

    r_squared = matrix_comparison(
        simmat_a, simmat_b, score=score, elements=elements
    )
    return r_squared


def rotation_matrix(theta):
    """Return 2D rotation matrix.

    Arguments:
        theta: Scalar value indicating radians of rotation.

    """
    return np.array((
        (np.cos(theta), -np.sin(theta)),
        (np.sin(theta), np.cos(theta)),
    ))


def affine_transformation(z, params):
    """Return affine transformation of 2D points.

    Arguments:
        z: Original set of points.
            shape = (n_point, 2)
        params: Transformation parameters denoting x translation,
            y translation, rotation (in radians), scaling, and flip
            factor.
            shape = (5,)

    """
    x = params[0]
    y = params[1]
    translation = np.array((x, y))
    theta = params[2]
    s = params[3]
    f = params[4]
    x = f * x
    r = rotation_matrix(theta)
    z_trans = s * (np.matmul(z, r) + translation)
    return z_trans


def procrustean_solution(z_a, z_b, n_restart=10):
    """Align the two embeddings using an affine transformation.

    Arguments:
        z_a: A 2D embedding.
            shape = (n_point, 2)
        z_b: A 2D embedding.
            shape = (n_point, 2)
        n_restart (optional): A scalar indicating the number of
            restarts for the optimization routine.

    Returns:
        z_c: An affine transformation of z_b that is maximally aligned
            with z_a.
        params: The affine transformation parameters.

    """
    def objective_fn(params, z_a, z_b):
        z_c = affine_transformation(z_b, params)
        # Loss is defined as the MSE of L2 distance.
        loss = np.mean(np.sum((z_a - z_c)**2, axis=1))
        return loss

    params_best = np.array((0., 0., 0., 1.))
    loss_best = np.inf
    for _ in range(n_restart):
        (x0, y0) = np.random.rand(2)
        theta0 = 2 * np.pi * np.random.rand(1)
        s0 = 2 * np.random.rand(1)
        # Perform a flip on some restarts.
        f0 = 1.
        if np.random.rand(1) < .5:
            f0 = -1.
        params0 = np.array((x0, y0, theta0, s0, f0))
        bnds = (
            (None, None),
            (None, None),
            (0., 2*np.pi),
            (0., None),
            (f0, f0)
        )
        res = minimize(objective_fn, params0, args=(z_a, z_b), bounds=bnds)
        params_candidate = res.x
        loss_candidate = res.fun
        if loss_candidate < loss_best:
            loss_best = loss_candidate
            params_best = params_candidate
        z_c = affine_transformation(z_b, params_best)
    return (z_c, params_best)


class ProgressBar(object):
    """Display a progress bar in terminal."""

    def __init__(
            self, total, prefix='', decimals=1, length=100,
            fill='█'):
        """Initialize.

        Arguments:
            iteration: Integer indicating current iteration.
            total: Integer indicating total iterations.
            prefix (optional): String that is used as prefix.
            suffix (optional): String that is used as suffix.
            decimals (optional): Integer indicating a positive number
                of decimals in percent complete.
            length (optional): Integer indicating the character length
                of the progress bar.
            fill (optional): String indicating the bar fill character.

        """
        self.total = total
        self.prefix = prefix
        self.decimals = decimals
        self.length = length
        self.fill = fill
        self.n_call = 0
        self.start_s = 0
        self.total_s = 0

    def _start(self):
        """Start timer."""
        self.start_s = time.time()

    def _stop(self):
        """Stop time."""
        self.total_s = time.time() - self.start_s

    def update(self, iteration):
        """Update progress bar to display current iteration."""
        # Start time if this is the first call.
        if self.n_call == 0:
            self._start()
        self.n_call = self.n_call + 1

        percent = ("{0:." + str(self.decimals) + "f}").format(
            100 * (iteration / float(self.total))
        )

        elapsed_time = time.time() - self.start_s
        if iteration == 0:
            time_per_iter = 0.0
        else:
            time_per_iter = elapsed_time / iteration

        eta_s = np.round((self.total - iteration) * time_per_iter)
        ett_s = np.round(self.total * time_per_iter)

        eta_str = str(datetime.timedelta(seconds=eta_s))
        ett_str = str(datetime.timedelta(seconds=ett_s))

        filledLength = int(self.length * iteration // self.total)
        bar = self.fill * filledLength + '-' * (self.length - filledLength)
        print(
            '\r    {0} |{1}| {2}% | ETA: {3} | ETT: {4}'.format(
                self.prefix, bar, percent, eta_str, ett_str
            ), end='\r'
        )
        # Print a new line on completion.
        if iteration == self.total:
            self._stop()
            print()



# -*- coding: utf-8 -*-
# Copyright 2020 The PsiZ Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for generating unjudged similarity judgment trials.

Classes:
    DocketGenerator: Base class for generating a docket of unjudged
        similarity trials.
    RandomGenerator: Concrete class for generating random similarity
        trials.
    ActiveGenerator: Concrete class for generating similarity trials
        using an active selection procedure that leverages expected
        information gain.

Functions:
    information_gain: Compute expected information gain of a docket.
    query_kl_priority: Compute query stimulus priority using KL
        divergence approach for fitted Gaussian.
    normal_kl_divergence: Compute KL divergence for two multi-variate
        Gaussian distributions.
    stimulus_entropy: Compute entropy of stimulus based on Gaussian
        approximation to posterior samples.
    normal_entropy: Compute entropy associated with a Gaussian
        distribution.
    choice_wo_replace: Efficient sampling without replacement.

Todo:
    - MAYBE Reduce uncertainty on positions AND group-specific
        parameters.

"""

from abc import ABCMeta, abstractmethod
import copy
# from functools import partial
import itertools
# import multiprocessing
import time

import numpy as np
import numpy.ma as ma
import pandas as pd
from sklearn.mixture import GaussianMixture
from sklearn.neighbors import NearestNeighbors

from psiz.trials import Docket, stack
from psiz.simulate import Agent
from psiz.preprocess import remove_catch_trials
from psiz.utils import ProgressBar


class DocketGenerator(object):
    """Abstract base class for generating similarity judgment trials.

    Methods:
        generate: Generate unjudged trials.

    """

    __metaclass__ = ABCMeta

    def __init__(self):
        """Initialize."""

    @abstractmethod
    def generate(self, args):
        """Return generated trials based on provided arguments.

        Arguments:
            n_stimuli

        Returns:
            A Docket object.

        """
        pass


class RandomGenerator(DocketGenerator):
    """A trial generator that independently samples trials."""

    def __init__(self, n_stimuli, n_reference=2, n_select=1, is_ranked=True):
        """Initialize.

        Arguments:
            n_stimuli: A scalar indicating the total number of unique
                stimuli.
            n_reference (optional): A scalar indicating the number of
                references for each trial.
            n_select (optional): A scalar indicating the number of
                selections an agent must make.
            is_ranked (optional): Boolean indicating whether an agent
                must make ranked selections.

        """
        DocketGenerator.__init__(self)

        self.n_stimuli = n_stimuli

        # Sanitize inputs.
        # TODO re-use sanitize methods from elsewhere
        self.n_reference = np.int32(n_reference)
        self.n_select = np.int32(n_select)
        self.is_ranked = bool(is_ranked)

    def generate(self, n_trial):
        """Return generated trials based on provided arguments.

        Arguments:
            n_trial: A scalar indicating the number of trials to
                generate.

        Returns:
            A Docket object.

        """
        n_reference = self.n_reference
        n_select = np.repeat(self.n_select, n_trial)
        is_ranked = np.repeat(self.is_ranked, n_trial)
        idx_eligable = np.arange(self.n_stimuli, dtype=np.int32)
        prob = np.ones([self.n_stimuli]) / self.n_stimuli
        stimulus_set = choice_wo_replace(
            idx_eligable, (n_trial, n_reference + 1), prob
        )

        return Docket(
            stimulus_set, n_select=n_select, is_ranked=is_ranked
        )


class ActiveGenerator(DocketGenerator):
    """A trial generator that leverages expected information gain.

    Since the number of candidate trials is typically enormous, a
    number of heuristics are used to guide the selection processes
    and narrow the search space.

    Attributes:
        n_stimuli:
        max_query:
        max_neighbor:
        max_candidate:

    Methods:
        generate:

    """

    def __init__(
            self, n_stimuli, n_reference=2, n_select=1, is_ranked=True,
            max_query=None, max_neighbor=1000, max_candidate=1000):
        """Initialize.

        Arguments:
            n_stimuli: A scalar indicating the total number of
                unique stimuli.
            n_reference (optional): A scalar indicating the number of
                references for each trial.
            n_select (optional): A scalar indicating the number of
                selections an agent must make.
            is_ranked (optional): Boolean indicating whether an agent
            max_query (optional): A scalar parameter that governs
                heuristic behavior. The value indicates the maximum
                number of unique query stimuli that should be chosen.
                By default, this is equal `n_stimuli` (i.e., no
                heuristic used).
            max_neighbor (optional): A scalar parameter that governs
                heuristic behavior. When selecting references, this
                value determines which stimuli can be selected as
                references for a particular query stimulus. Only the
                stimuli that are no more than max_neighbor away will be
                considered. Increasing the value above the default is
                likely to increase computational time. It is not
                recommended to use a value smaller than the default.
            max_candidate (optional): A scalar parameter that
                governs heuristic behavior. Given a query stimulus,
                this parameter determines how many candidate trials
                will be considered. In general, more is better, but
                you may be limited by time and RAM. Must be greater
                than zero.

        """
        DocketGenerator.__init__(self)

        self.n_stimuli = n_stimuli

        # Set trial configuration parameters.
        self.n_reference = np.int32(n_reference)
        self.n_select = np.int32(n_select)
        self.is_ranked = bool(is_ranked)

        # Set heuristic parameters.
        if max_query is None:
            max_query = n_stimuli
        self.max_query = np.minimum(max_query, n_stimuli)
        self.max_neighbor = np.minimum(n_stimuli - 1, max_neighbor)
        self.max_candidate = max_candidate

    def generate(self, n_trial, embedding, samples, priority, verbose=0):
        """Return a docket of trials based on provided arguments.

        Trials are selected in order to maximize expected information
        gain given a specific group. Expected information gain is
        approximated using MCMC posterior samples.

        The docket is assembled in two steps, first the query stimuli
        are chosen. The values in `priority` are used to select query
        stimuli that the user considers promising. Second, reference
        stimuli for the queries are chosen. References are selected by
        only considering nearby neighbors.

        Arguments:
            n_trial: A scalar indicating the number of trials to
                generate.
            embedding: A PsychologicalEmbedding object.
            samples: A dictionary containing the posterior
                samples of parameters from a PsychologicalEmbedding
                object. The samples should correspond to group_id=0 in
                the `embedding` object.
            priority: An array indicating the priority of
                each stimulus. Priority is used as a heuristic to guide
                the active selection procedure in the typically vast
                search space.
            verbose (optional): An integer specifying the verbosity of
                printed output. If zero, nothing is printed. Increasing
                integers display an increasing amount of information.

        Returns:
            docket: A Docket object with trials that maximize expected
                information gain.
            info: A dictionary containing additional information about
                the docket.
                ig_trial: A numpy.ndarray containing the expected
                information gain for each trial in the docket.
                shape = (n_trial,)

        Notes:
            The variable `samples` must be samples from a particular
                group. It is assumed that group_id=0 in `embedding`
                was used to generate these samples.

        """
        if verbose > 0:
            print('[psiz] Generating docket using active selection...')


        # Ensure priorities sum to one. TODO
        # if np.sum(np.less(priority, 0)) > 0:
        #     raise('ValueError')
        # priority = priority - np.min(priority)
        # priority -= np.min(priority)
        priority = priority / np.sum(priority)

        # Determine number of unique query stimuli in docket.
        n_query = np.minimum(n_trial, self.max_query)

        # Assemble docket in two stages.
        (query_idx_list, n_trial_per_query_list) = self._select_query(
            n_trial, priority, n_query
        )
        (docket, ig) = self._select_references(
            embedding, samples, query_idx_list, n_trial_per_query_list,
            verbose=verbose
        )
        info = {
            "ig_trial": ig
        }

        return (docket, info)

    def _select_query(self, n_trial, priority, n_query):
        """Select which stimuli should by queries and how often.

        Arguments:
            priority:
            n_query: Scalar indicating the number of unique queries.

        Returns:
            query_idx_list: An array of selected query indices.
            n_trial_per_query_list: An array indicating the corresponding
                number of times a query should be shown in a trial.

        """
        n_stimuli = len(priority)

        # Create an index list for all selected query stimuli.
        # Initialize index with all stimuli.
        query_idx_list = np.arange(0, n_stimuli)

        # If necessary, stochastically select subset of query stimuli.
        if n_query < n_stimuli:
            query_idx_list = np.random.choice(
                query_idx_list, n_query, replace=False, p=priority
            )

        # Determine how many times each query stimulus should be used.
        n_trial_per_query_list = np.zeros((n_query), dtype=np.int32)
        for i_trial in range(n_trial):
            n_trial_per_query_list[np.mod(i_trial, n_query)] = (
                n_trial_per_query_list[np.mod(i_trial, n_query)] + 1
            )

        return query_idx_list, n_trial_per_query_list

    def _select_references(
            self, embedding, samples, query_idx_list, n_trial_per_query_list,
            verbose=0):
        """Determine references for all requested query stimuli.

        Notes:
            This function assumes that group_id=0 of the embedding
                should be used.

        """
        n_query = query_idx_list.shape[0]

        if verbose > 0:
            progbar = ProgressBar(
                n_query, prefix='Progress:', length=50
            )
            progbar.update(0)

        best_docket = None
        ig_best = None
        for i_query in range(n_query):
            top_candidate, curr_best_ig = _select_query_references(
                i_query, embedding, samples, query_idx_list,
                n_trial_per_query_list,
                self.n_reference, self.n_select, self.is_ranked,
                self.max_candidate, self.max_neighbor
            )

            if verbose > 0:
                progbar.update(i_query + 1)

            # Add to dynamic list.
            if ig_best is None:
                ig_best = curr_best_ig
            else:
                ig_best = np.hstack((ig_best, curr_best_ig))

            if best_docket is None:
                best_docket = top_candidate
            else:
                best_docket = stack((best_docket, top_candidate))

        return (best_docket, ig_best)


def _select_query_references(
        i_query, embedding, samples, query_idx_list, n_trial_per_query_list,
        n_reference, n_select, is_ranked, max_candidate, max_neighbor):
    """Determine query references.

    Arguments:
        max_neighbor (optional): Scalar indicating the number of query
            neighbors to consider as references. This parameter
            primarily acts as a bound on the computational cost of the
            sampling without replacement function.

    Notes:
        This function assumes that group_id=0 of the embedding
            should be used.

    """
    query_idx = query_idx_list[i_query]
    n_trial_q = n_trial_per_query_list[i_query]

    n_stimuli = embedding.n_stimuli
    z = embedding.z
    z_q = z[query_idx, :]
    s_qr = embedding.similarity(np.expand_dims(z_q, axis=0), z)

    # Determine eligable reference stimuli and their draw probability
    # based on similarity. TODO remove self
    ref_idx_eligable = np.argsort(-s_qr)
    # Remove query from eligable list. Note that we don't just pop the
    # first element since the user may be using an exotic similarity
    # function.
    ref_idx_eligable = ref_idx_eligable[
        np.not_equal(ref_idx_eligable, query_idx)
    ]
    # Limit eligable references to nearest neighbors.
    ref_idx_eligable = ref_idx_eligable[0:max_neighbor]
    # Renormalize probability of drawing a reference based on similarity
    # of the eligable options.
    ref_prob = s_qr[ref_idx_eligable] / np.sum(s_qr[ref_idx_eligable])

    # Create a docket full of candidate trials.
    n_select = np.repeat(n_select, max_candidate)
    is_ranked = np.repeat(is_ranked, max_candidate)
    stimulus_set = np.empty(
        (max_candidate, n_reference + 1), dtype=np.int32
    )
    stimulus_set[:, 0] = query_idx
    stimulus_set[:, 1:] = choice_wo_replace(
        ref_idx_eligable, (max_candidate, n_reference), ref_prob
    )
    docket = Docket(
        stimulus_set, n_select=n_select, is_ranked=is_ranked
    )

    # Compute information gain of candidate trials.
    ig = information_gain(embedding, samples, docket)

    # Grab the top trials as requested.
    top_indices = np.argsort(-ig)
    docket_top = docket.subset(
        top_indices[0:n_trial_q]
    )
    ig_top = ig[top_indices[0:n_trial_q]]

    return docket_top, ig_top


def information_gain(embedding, samples, docket, group_id=None):
    """Return expected information gain of trial(s) in docket.

    Information gain is determined by computing the mutual
    mutual information between the candidate trial(s) and the
    existing set of observations.

    Arguments:
        embedding: A PsychologicalEmbedding object.
        samples: Dictionary of sampled parameters.
            'z': shape = (n_stimuli, n_dim, n_sample)
        docket: A Docket object.
        group_id (optional): A scalar or an array with
            shape = (n_trial,).

    Returns:
        A numpy.ndarray representing the expected information gain
        of the candidate trial(s).
        shape = (n_trial,)

    """
    cap = 2.2204e-16

    # Note: z_samples has shape = (n_stimuli, n_dim, n_sample)
    z_samples = samples['z']
    # Note: prob_all has shape = (n_trial, n_outcome, n_sample)
    prob_all = embedding.outcome_probability(
        docket, group_id=group_id, z=z_samples
    )

    # First term of mutual information.
    # H(Y | obs, c) = - sum P(y_i | obs, c) log P(y_i | obs, c)
    # Take mean over samples to approximate p(y_i | obs, c).
    # shape = (n_trial, n_outcome)
    first_term = ma.mean(prob_all, axis=2)
    # Use threshold to avoid log(0) issues (unlikely to happen).
    first_term = ma.maximum(cap, first_term)
    first_term = first_term * ma.log(first_term)
    # Sum over possible outcomes.
    first_term = -1 * ma.sum(first_term, axis=1)

    # Second term of mutual information.
    # E[H(Y | Z, D, x)]
    # Use threshold to avoid log(0) issues (likely to happen).
    # shape = (n_trial, n_outcome, n_sample)
    prob_all = ma.maximum(cap, prob_all)

    # shape = (n_trial, n_outcome, n_sample)
    second_term = prob_all * ma.log(prob_all)

    # Take the sum over the possible outcomes.
    # shape = (n_trial, n_sample)
    second_term = ma.sum(second_term, axis=1)

    # Take the sum over all samples.
    # shape = (n_trial)
    second_term = ma.mean(second_term, axis=1)

    info_gain = first_term + second_term

    # Convert to normal numpy.ndarray.
    info_gain = info_gain.data
    return info_gain


def choice_wo_replace(a, size, p):
    """Fast sampling without replacement.

    Arguments:
        a: An array indicating the eligable elements.
        size: A tuple indicating the number of independent samples and
            the number of draws (without replacement) for each sample.
            The tuple is ordered such that
            size = (n_sample, sample_size).
        p: An array indicating the probabilites associated with drawing
            a particular element. User provided probabilities are
            already assumed to sum to one. Probability p[i] indicates
            the probability of drawing index a[i].

    Returns:
        result: A 2D array containing the drawn elements.
            shape=(n_sample, sample_size)

    See: https://medium.com/ibm-watson/
        incredibly-fast-random-sampling-in-python-baf154bd836a

    """
    n_sample = size[0]
    sample_size = size[1]

    # Replicate probabilities as many times as `n_sample`
    replicated_probabilities = np.tile(p, (n_sample, 1))

    # Get random shifting numbers & scale them correctly.
    random_shifts = np.random.random(replicated_probabilities.shape)
    random_shifts /= random_shifts.sum(axis=1)[:, np.newaxis]

    # Shift by numbers and find largest (by finding the smallest of the
    # negative).
    shifted_probabilities = random_shifts - replicated_probabilities
    samples = np.argpartition(
        shifted_probabilities, sample_size, axis=1
    )[:, :sample_size]

    return a[samples]


def determine_stimulus_priority(embedding, samples, mode='kl'):
    """Determine stimulus priority.

    The priority of all stimuli must sum to one.

    Arguments:
        embedding:
        samples:
        mode (optional): Can be 'random', 'entropy' or 'kl'.

    Returns:
        priority:

    """
    n_stimuli = embedding.n_stimuli

    if mode is "random":
        # Completely random.
        priority = np.ones([n_stimuli]) / n_stimuli
    elif mode is "entropy":
        # Based on entropy.
        entropy = stimulus_entropy(samples)
        rel_entropy = entropy - np.min(entropy)
        priority = rel_entropy / np.sum(rel_entropy)
    elif mode is "kl":
        # Based on KL divergence.
        stim_kl = query_kl_priority(embedding, samples)
        priority = stim_kl / np.sum(stim_kl)

    return priority


def stimulus_entropy(samples):
    """Return the approximate entropy associated with every stimulus.

    Arguments:
        samples: Posterior samples.
            shape: (n_stimuli, n_dim, n_sample)

    Returns:
        The approximate entropy associated with each stimulus.
            shape: (n_stimuli,)

    Notes:
        The computation is specific to a particular group.

    """
    # Unpack.
    z_samp = samples['z']
    n_stimuli = z_samp.shape[0]

    # Fit multi-variate normal for each stimulus.
    z_samp = np.transpose(z_samp, axes=[2, 0, 1])
    entropy = np.empty((n_stimuli))
    for i_stim in range(n_stimuli):
        gmm = GaussianMixture(
            n_components=1, covariance_type='full'
        )
        gmm.fit(z_samp[:, i_stim, :])
        entropy[i_stim] = normal_entropy(gmm.covariances_[0])
    return entropy


def normal_entropy(cov):
    """Return entropy of multivariate normal distribution."""
    n_dim = cov.shape[0]
    h = (
        (n_dim / 2) +
        (n_dim / 2 * np.log(2 * np.pi)) +
        (1 / 2 * np.log(np.linalg.det(cov)))
    )
    return h


def query_kl_priority(embedding, samples):
    """Return a priority score for every stimulus.

    The score indicates the priority each stimulus should serve as
    a query stimulus in a trial.

    Arguments:
        embedding:
        samples:

    Returns:
        A priority score.
            shape: (n_stimuli,)

    Notes:
        The priority scores are specific to a particular group.

    """
    # Unpack.
    # z = embedding.z
    z = np.median(samples['z'], axis=2)
    try:
        rho = embedding.rho
    except AttributeError:
        rho = 2.
    z_samp = samples['z']
    # z_median = np.median(z_samp, axis=2)
    (n_stimuli, n_dim, _) = z_samp.shape

    # Fit multi-variate normal for each stimulus.
    z_samp = np.transpose(z_samp, axes=[2, 0, 1])
    mu = np.empty((n_stimuli, n_dim))
    cov = np.empty((n_stimuli, n_dim, n_dim))
    for i_stim in range(n_stimuli):
        gmm = GaussianMixture(
            n_components=1, covariance_type='full'
        )
        gmm.fit(z_samp[:, i_stim, :])
        mu[i_stim, :] = gmm.means_[0]
        cov[i_stim, :, :] = gmm.covariances_[0]

    # Determine nearest neighbors.
    # TODO maybe convert to faiss
    k = np.minimum(10, n_stimuli-1)
    nbrs = NearestNeighbors(
        n_neighbors=k+1, algorithm='auto', p=rho
    ).fit(z)
    (_, nn_idx) = nbrs.kneighbors(z)
    nn_idx = nn_idx[:, 1:]  # Drop self index.

    # Compute KL divergence for nearest neighbors.
    kl_div = np.empty((n_stimuli, k))
    for query_idx in range(n_stimuli):
        for j_nn in range(k):
            idx_j = nn_idx[query_idx, j_nn]
            kl_div[query_idx, j_nn] = normal_kl_divergence(
                mu[idx_j], cov[idx_j], mu[query_idx], cov[query_idx]
            )
    score = np.sum(kl_div, axis=1)
    score = 1 / score

    return score


def normal_kl_divergence(mu_a, cov_a, mu_b, cov_b):
    """Return the Kullback-Leibler divergence between two normals."""
    mu_diff = mu_b - mu_a
    mu_diff = np.expand_dims(mu_diff, 1)
    n_dim = len(mu_a)
    cov_b_inv = np.linalg.inv(cov_b)
    kl = .5 * (
        np.log(np.linalg.det(cov_b) / np.linalg.det(cov_a)) - n_dim +
        np.matrix.trace(np.matmul(cov_b_inv, cov_a)) +
        np.matmul(
            np.matmul(np.transpose(mu_diff), cov_b_inv), mu_diff
        )
    )
    return kl



# -*- coding: utf-8 -*-
# Copyright 2020 The PsiZ Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for preprocessing observations.

Functions:
    identify_catch_trials: Identify catch trials.
    grade_catch_trials: Identify and grade catch trials.
    quality_control: Remove observations belonging to agents that do
        not meet quality control standards.
    remove_catch_trials: Remove catch trials.

"""

from pathlib import Path

import numpy as np
import pandas as pd


def identify_catch_trials(obs):
    """Identify catch trials.

    Catch trials are assumed to be any trial where at least one of the
    references is the same stimulus as the query.

    Arguments:
        obs: A psiz.trials.Observations object.

    Returns:
        is_catch: Boolean array indicating catch trial locations.
            shape = (n_trial,)

    """
    n_trial = obs.n_trial
    is_catch = np.zeros([n_trial], dtype=bool)
    for i_trial in range(n_trial):
        # Determine which references are identical to the query.
        is_identical = np.equal(
            obs.stimulus_set[i_trial, 0], obs.stimulus_set[i_trial, 1:]
        )
        if np.sum(is_identical) > 0:
            is_catch[i_trial] = True
    return is_catch


def grade_catch_trials(obs, grade_mode='lenient'):
    """Grade catch trials.

    Catch trials are assumed to be any trial where at least one of the
    references is the same stimulus as the query. A catch trial is
    graded as correct depending on the grade_mode.

    Arguments:
        obs: A psiz.trials.Observations object.
        grade_mode (optional): Determines the manner in which responses
            are graded. Can be either 'strict', 'partial', or
            'lenient'. The options 'strict' and 'partial' are only
            relevant for trials where participants provide ranked
            responses, otherwise they are equivalent to 'lenient'. If
            'lenient', then one of the selected references must include
            the copy of the query. If 'strict', then the first choice
            must be the copy of the query. If 'partial', then full
            credit is given if the first choice is the copy of the
            query and half credit is given if a choice other than the
            first choice includes a copy of the query.

    Returns:
        avg_grade: Scalar indicating average grade on all catch trials.
            Is np.nan if there are no catch trials.
        grade: Array indicating grade of catch trial. The value can be
            between 0 and 1, where 1 is a perfect score.
            shape = (n_catch_trial,)
        is_catch: Boolean array indicating catch trial locations.
            shape = (n_trial,)

    """
    n_trial = obs.n_trial
    is_catch = identify_catch_trials(obs)
    grade = np.zeros([n_trial])

    for i_trial in range(n_trial):
        if is_catch[i_trial]:
            # Determine which references are identical to the query.
            is_identical = np.equal(
                obs.stimulus_set[i_trial, 0], obs.stimulus_set[i_trial, 1:]
            )
            # Grade response.
            if grade_mode is 'lenient':
                is_identical_selected = is_identical[0:obs.n_select[i_trial]]
                if np.sum(is_identical_selected) > 0:
                    grade[i_trial] = 1
            elif grade_mode is 'strict':
                if obs.is_ranked[i_trial]:
                    is_identical_selected = is_identical[0]
                else:
                    is_identical_selected = is_identical[
                        0:obs.n_select[i_trial]
                    ]
                if np.sum(is_identical_selected) > 0:
                    grade[i_trial] = 1
            elif grade_mode is 'partial':
                if obs.is_ranked[i_trial]:
                    is_identical_selected = is_identical[0]
                    if is_identical_selected:
                        grade[i_trial] = 1
                    else:
                        is_identical_selected = is_identical[
                            0:obs.n_select[i_trial]
                        ]
                        if np.sum(is_identical_selected) > 0:
                            grade[i_trial] = .5
                else:
                    is_identical_selected = is_identical[
                        0:obs.n_select[i_trial]
                    ]
                    if np.sum(is_identical_selected) > 0:
                        grade[i_trial] = 1
            else:
                raise ValueError((
                    "The argument `grade_mode` must be 'strict', 'partial' or"
                    " 'lenient'."
                ))

    # Compute average grade.
    grade = grade[is_catch]
    if len(grade) > 0:
        n_catch = np.sum(is_catch)
        avg_grade = np.sum(grade) / n_catch
    else:
        avg_grade = np.nan

    return (avg_grade, grade, is_catch)


def quality_control(obs, grade_thresh=1.0, grade_mode='lenient'):
    """Remove agents that do not meet quality control standards.

    Arguments:
        obs: A psiz.trials.Observations object.
        grade_thresh (optional): The threshold for dropping
            an agent's data.
        grade_mode (optional): Determines the manner in which responses
            are graded. See function grade_catch_trials.

    Returns:
        obs: An psiz.trials.Observations object with bad data removed.

    """
    agent_list = np.unique(obs.agent_id)
    grade_record = {
        'agent_id': agent_list,
        'grade': np.zeros(len(agent_list)),
        'is_retained': np.ones(len(agent_list), dtype=bool)
    }

    keep_locs = np.ones([obs.n_trial], dtype=bool)
    for idx, i_agent in enumerate(agent_list):
        agent_locs = np.equal(obs.agent_id, i_agent)
        obs_agent = obs.subset(agent_locs)
        (avg_grade, _, _) = grade_catch_trials(
            obs_agent, grade_mode=grade_mode
        )
        grade_record['grade'][idx] = avg_grade

        # Drop agent if below required grade threshold.
        # Note that this retains observations with zero catch trials.
        if avg_grade < grade_thresh:
            keep_locs[agent_locs] = False
            grade_record['is_retained'][idx] = False

    obs_new = obs.subset(keep_locs)
    df_grade = pd.DataFrame.from_dict(grade_record)
    return obs_new, df_grade


def remove_catch_trials(obs):
    """Remove all catch trials.

    Arguments:
        obs: A psiz.trials.Observations object.

    Returns:
        obs: A psiz.trials.Observations object.

    """
    is_catch = identify_catch_trials(obs)
    obs = obs.subset(np.logical_not(is_catch))
    return obs



"""Setup file."""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='psiz',
    version='0.3.0',
    description='Toolbox for inferring psychological embeddings.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    author='Brett D. Roads',
    author_email='brett.roads@gmail.com',
    license='Apache Licence 2.0',
    packages=['psiz'],
    python_requires='>=3, <3.8',
    install_requires=[
        'tensorflow>=2', 'scipy', 'pandas', 'scikit-learn',
        'matplotlib', 'tensorflow-probability', 'numexpr', 'pillow'
    ],
    include_package_data=True,
    url='https://github.com/roads/psiz',
    download_url='https://github.com/roads/psiz/archive/v0.3.0.tar.gz'
)



# -*- coding: utf-8 -*-
# Copyright 2020 The PsiZ Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for selecting the dimensionality of an embedding.

Functions:
    search: Use a validation procedure to select the dimensionality for
        an embedding procedure.
    dimension_search: Use a validation procedure to select the
        dimensionality for an embedding procedure. DEPRECATED, use
        search instead.
    visualize_dimension_search: Visualize results of dimension search.

"""

import os

import numpy as np
from sklearn.model_selection import StratifiedKFold
import matplotlib.pyplot as plt
from scipy.stats import sem

import psiz.utils as ut


def search(obs, model_spec, search_spec=None, verbose=0):
    """Search for an appropriate embedding dimensionality.

    Search over the list of candidate dimensions, starting with the
    smallest, in order to find the best dimensionality for the data.
    Dimensions are examined in ascending order. The search stops when
    adding dimensions does not reduce loss or there are no more
    dimensions in the search list. Each dimension is evaluated using
    the same cross-validation partitions in order to make comparisons
    as equitable as possible.

    Arguments:
        obs: An Observations object representing the observed data.
            embedding_constructor: A PsychologicalEmbedding
            constructor.
        model_spec: A dictionary specifying the embedding model to use.
        search_spec (optional): A dictionary specifying the parameters
            of the search procedure.
        verbose (optional): An integer specifying the verbosity of
            printed output.

        model_spec: TODO
        n_stimuli:  An integer indicating the number of unique stimuli.
        n_restart (optional): An integer specifying the number of
            restarts to use for the inference procedure. Since the
            embedding procedure finds local optima, multiple restarts
            helps find the global optimum.
        modifier_func (optional): A function that takes an embedding
            as the only argument and returns a modified embedding. This
            argument can be used to modify an embedding after it is
            initialized. For example, to set and freeze parameters.

        search_spec: TODO
        dim_list (optional): A list of integers indicating the
            dimensions to search over.
        n_split (optional): Integer specifying how many splits to
            create from the data. This defines the proportion of
            train and test data.
        n_fold (optional): Integer specifying the number of folds to
            use for cross-validation when selection the dimensionality.
            Must be at least one and cannot be more than n_split.
        max_patience (optional): Integer specifying how many dimensions
            to wait for an improvement in test loss.

    Returns:
        summary: A dictionary.
            dim_list: The dimensionalities searched.
            loss_train: The training loss.
                shape=(len(dim_list), n_fold)
            loss_test: The test loss.
                shape=(len(dim_list), n_fold)
            dim_best: An integer indicating the dimensionality
                (from the candidate list) that minimized loss on the
                held-out data.

    """
    if search_spec is None:
        search_spec = {
            'dim_list': range(2, 51),
            'n_restart': 100,
            'n_split': 10,
            'n_fold': 10
        }

    # Unpack and check.
    dim_list = np.sort(search_spec['dim_list'])
    n_restart = search_spec['n_restart']
    n_split = search_spec['n_split']
    n_fold = search_spec['n_fold']
    max_patience = search_spec['max_patience']
    # if search_spec['n_fold'] > search_spec['n_split']: TODO issue error

    if (verbose > 0):
        print('[psiz] Searching dimensionality ...')
        print('  Settings:')
        print('    Dimensionality search list: ', dim_list)
        print('    Splits: ', n_split)
        print('    Folds: ', n_fold)
        print('    Restarts per fold: ', n_restart)
        print('    Patience: ', max_patience)
        print('')

    # Instantiate the balanced k-fold cross-validation object.
    skf = StratifiedKFold(n_splits=n_split)
    split_list = list(
        skf.split(obs.stimulus_set, obs.group_id)
    )

    # Sweep over the list of candidate dimensions.
    loss_test_avg_best = np.inf
    loss_train = np.nan * np.ones([len(dim_list), n_fold])
    loss_val = np.nan * np.ones([len(dim_list), n_fold])
    loss_test = np.nan * np.ones([len(dim_list), n_fold])
    patience = 0
    for idx_dim, i_dimension in enumerate(dim_list):
        # Instantiate embedding.
        emb = model_spec['model'](
            model_spec['n_stimuli'], n_dim=i_dimension,
            n_group=model_spec['n_group']
        )
        if model_spec['modifier'] is not None:
            emb = model_spec['modifier'](emb)
        if verbose > 1:
            print('  Dimensionality: ', i_dimension)

        for i_fold in range(n_fold):
            (train_index, test_index) = split_list[i_fold]
            if verbose > 2:
                print('    Fold: ', i_fold)
            # Train
            obs_train = obs.subset(train_index)
            loss_train[idx_dim, i_fold], loss_val[idx_dim, i_fold] = emb.fit(
                obs_train, n_restart=n_restart, verbose=verbose-1
            )
            # Test
            obs_test = obs.subset(test_index)
            loss_test[idx_dim, i_fold] = emb.evaluate(obs_test)

            i_fold = i_fold + 1
        # Compute average cross-validation train and test loss.
        loss_train_avg = np.mean(
            (.9 * loss_train[idx_dim, :]) + (.1 * loss_val[idx_dim, :])
        )
        loss_test_avg = np.mean(loss_test[idx_dim, :])

        if verbose > 1:
            print("    Avg. Train Loss: {0:.2f}".format(loss_train_avg))
            print("    Avg. Test Loss: {0:.2f}".format(loss_test_avg))

        if loss_test_avg < loss_test_avg_best:
            # Larger dimensionality yielded a better test loss.
            loss_test_avg_best = loss_test_avg
            best_dimensionality = i_dimension
            patience = 0
            if verbose > 1:
                print("    Test loss improved.")
        else:
            # Larger dimensionality yielded a worse test loss.
            patience = patience + 1
            if verbose > 1:
                print(
                    "    Test loss did not improve."
                    "(patience={0})".format(patience)
                )
        if verbose > 1:
            print("")

        if patience > max_patience:
            # Stop search.
            break

    if verbose > 0:
        print('Best dimensionality: ', best_dimensionality)

    summary = {
        "dim_list": dim_list,
        "loss_train": loss_train,
        "loss_test": loss_test,
        "dim_best": best_dimensionality
    }
    return summary


def dimension_search(
        obs, embedding_constructor, n_stimuli, dim_list=None,
        modifier_func=None, n_restart=20, n_split=5, n_fold=1,
        max_patience=1, verbose=0):
    """Suggest an embedding dimensionality given provided observations.

    Search over the list of candidate dimensions, starting with the
    smallest, in order to find the best dimensionality for the data.
    Dimensions are examined in ascending order. The search stops when
    adding dimensions does not reduce loss or there are no more
    dimensions in the search list. Each dimension is evaluated using
    the same cross-validation partitions.

    Arguments:
        obs: An Observations object representing the observed data.
            embedding_constructor: A PsychologicalEmbedding
            constructor.
        n_stimuli:  An integer indicating the number of unique stimuli.
        dim_list (optional): A list of integers indicating the
            dimensions to search over.
        modifier_func (optional): A function that takes an embedding
            as the only argument and returns a modified embedding. This
            argument can be used to modify an embedding after it is
            initialized. For example, to set and freeze parameters.
        n_restart (optional): An integer specifying the number of
            restarts to use for the inference procedure. Since the
            embedding procedure finds local optima, multiple restarts
            helps find the global optimum.
        n_split (optional): Integer specifying how many splits to
            create from the data. This defines the proportion of
            train and test data.
        n_fold (optional): Integer specifying the number of folds to
            use for cross-validation when selection the dimensionality.
            Must be at least one and cannot be more than n_split.
        max_patience (optional): Integer specifying how many dimensions
            to wait for an improvement in test loss.
        verbose (optional): An integer specifying the verbosity of
            printed output.

    Returns:
        summary: A dictionary.
            dim_best: An integer indicating the dimensionality
                (from the candidate list) that minimized the loss
                function.

    """
    n_group = len(np.unique(obs.group_id))

    if dim_list is None:
        dim_list = range(2, 51)
    else:
        # Make sure dimensions are in ascending order.
        dim_list = np.sort(dim_list)

    if (verbose > 0):
        print('Searching dimensionality ...')
        print('  Settings:')
        print('    Dimensionality search list: ', dim_list)
        print('    Splits: ', n_split)
        print('    Folds: ', n_fold)
        print('    Restarts per fold: ', n_restart)
        print('    Patience: ', max_patience)
        print('')

    # Instantiate the balanced k-fold cross-validation object.
    skf = StratifiedKFold(n_splits=n_split)
    split_list = list(
        skf.split(obs.stimulus_set, obs.group_id)
    )

    # Sweep over the list of candidate dimensions.
    loss_test_avg_best = np.inf
    loss_train = np.nan * np.ones([len(dim_list), n_fold])
    loss_test = np.nan * np.ones([len(dim_list), n_fold])
    patience = 0
    for idx_dim, i_dimension in enumerate(dim_list):
        # Instantiate embedding.
        emb = embedding_constructor(
            n_stimuli, n_dim=i_dimension, n_group=n_group
        )
        if modifier_func is not None:
            emb = modifier_func(emb)
        if verbose > 1:
            print('  Dimensionality: ', i_dimension)

        for i_fold in range(n_fold):
            (train_index, test_index) = split_list[i_fold]
            if verbose > 2:
                print('    Fold: ', i_fold)
            # Train
            obs_train = obs.subset(train_index)
            loss_train[idx_dim, i_fold], loss_val[idx_dim, i_fold] = emb.fit(
                obs_train, n_restart=n_restart, verbose=verbose-1
            )
            # Test
            obs_test = obs.subset(test_index)
            loss_test[idx_dim, i_fold] = emb.evaluate(obs_test)

            i_fold = i_fold + 1
        # Compute average cross-validation train and test loss.
        loss_train_avg = np.mean(loss_train[idx_dim, :])
        loss_test_avg = np.mean(loss_test[idx_dim, :])

        if verbose > 1:
            print("    Avg. Train Loss: {0:.2f}".format(loss_train_avg))
            print("    Avg. Test Loss: {0:.2f}".format(loss_test_avg))

        if loss_test_avg < loss_test_avg_best:
            # Larger dimensionality yielded a better test loss.
            loss_test_avg_best = loss_test_avg
            best_dimensionality = i_dimension
            patience = 0
            if verbose > 1:
                print("    Test loss improved.")
        else:
            # Larger dimensionality yielded a worse test loss.
            patience = patience + 1
            if verbose > 1:
                print(
                    "    Test loss did not improve."
                    "(patience={0})".format(patience)
                )
        if verbose > 1:
            print("")

        if patience > max_patience:
            # Stop search.
            break

    if verbose > 0:
        print('Best dimensionality: ', best_dimensionality)

    summary = {
        "dim_list": dim_list,
        "loss_train": loss_train,
        "loss_test": loss_test,
        "dim_best": best_dimensionality
    }
    return summary


def visualize_dimension_search(ax, summary):
    """Visualize dimensionality search.

    Arguments:
        ax: A Matplotlib axis

    Example usage:
        fig, ax = plt.subplots(figsize=(4, 3))
        ax = plt.subplot(1, 1, 1)
        visualize_dimension_search(ax, dim_summary)

    """
    dim_list = summary["dim_list"]
    dim_best = summary["dim_best"]

    train_mean = np.mean(summary["loss_train"], axis=1)
    test_mean = np.mean(summary["loss_test"], axis=1)

    ax.plot(dim_list, train_mean, 'o-b', markersize=3, label="Train")
    ax.plot(dim_list, test_mean, 'o-r', markersize=3, label="Test")

    if summary["loss_train"].shape[1] > 1:
        train_sem = sem(summary["loss_train"], axis=1)
        test_sem = sem(summary["loss_test"], axis=1)
        ax.fill_between(
            dim_list, train_mean - train_sem, train_mean + train_sem,
            alpha=.5
        )
        ax.fill_between(
            dim_list, test_mean - test_sem, test_mean + test_sem,
            alpha=.5
        )
    ax.scatter(
        dim_best, test_mean[np.equal(dim_list, dim_best)], c="r", s=50,
        marker='x', label='Best Dimensionality'
    )

    ax.set_xlabel("Dimensionality")
    ax.set_ylabel("Loss")
    ax.legend()
    ax.set_title('Dimensionality Search\n(Mean and SEM)')



# -*- coding: utf-8 -*-
# Copyright 2020 The PsiZ Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Module for similarity judgment trials.

Classes:
    SimilarityTrials: Abstract class for similarity judgment trials.
    Docket: Unjudged similarity judgment trials.
    Observations: Similarity judgment trials that have been judged and
        will serve as observed data during inference.

Functions:
    stack: Combine a list of multiple SimilarityTrial objects into one.
    squeeze: Squeeze indices to be small and consecutive.
    load: Load a hdf5 file, saved using the `save` class method,
        as a SimilarityTrial object.
    load_trials: DEPRECATED Alias for load.

Notes:
    On each similarity judgment trial, an agent judges the similarity
        between a single query stimulus and multiple reference stimuli.
    Groups are used to identify distinct populations of agents. For
        example, similarity judgments could be collected from two
        groups: novices and experts. During inference, group
        information can be used to infer a separate set of attention
        weights for each group while sharing all other parameters.

Todo:
    - MAYBE restructure group_id and agent_id. If we wanted to allow
    for arbitrary hierarchical models, maybe better off making
    group_id a 2D array of shape=(n_trial, n_group_level)
    - MAYBE make config_list a custom object

"""

from abc import ABCMeta, abstractmethod
from itertools import permutations
import copy
import warnings

import h5py
import numpy as np
import pandas as pd


class SimilarityTrials(object):
    """Abstract base class for similarity judgment trials.

    This abstract base class is used to organize data associated with
    similarity judgment trials. As the class name suggests, this object
    handles data associated with multiple trials. Depending on the
    concrete subclass, the similarity trials represent unjudged trials
    or judged trials.

    Attributes:
        n_trial: An integer indicating the number of trials.
        stimulus_set: An integer matrix containing indices that
            indicate the set of stimuli used in each trial. Each row
            indicates the stimuli used in one trial. The first column
            is the query stimulus. The remaining, columns indicate
            reference stimuli. Negative integers are used as
            placeholders to indicate non-existent references.
            shape = (n_trial, max(n_reference) + 1)
        n_reference: An integer array indicating the number of
            references in each trial.
            shape = (n_trial,)
        n_select: An integer array indicating the number of references
            selected in each trial.
            shape = (n_trial,)
        is_ranked: A Boolean array indicating which trials require
            reference selections to be ranked.
            shape = (n_trial,)
        config_idx: An integer array indicating the
            configuration of each trial. The integer is an index
            referencing the row of config_list and the element of
            outcome_idx_list.
            shape = (n_trial,)
        config_list: A DataFrame object describing the unique trial
            configurations.
        outcome_idx_list: A list of 2D arrays indicating all possible
            outcomes for a trial configuration. Each element in the
            list corresponds to a trial configuration in config_list.
            Each row of the 2D array indicates one potential outcome.
            The values in the rows are the indices of the the reference
            stimuli (as specified in the attribute `stimulus_set`.

    Methods:
        subset: Return a subset of similarity trials given an index.

    """

    __metaclass__ = ABCMeta

    def __init__(self, stimulus_set, n_select=None, is_ranked=None):
        """Initialize.

        Arguments:
            stimulus_set: An integer matrix containing indices that
                indicate the set of stimuli used in each trial. Each
                row indicates the stimuli used in one trial. The first
                column is the query stimulus. The remaining, columns
                indicate reference stimuli. It is assumed that stimuli
                indices are composed of integers from [0, N-1], where N
                is the number of unique stimuli. Negative integers are
                used as placeholders to indicate non-existent
                references.
                shape = (n_trial, max(n_reference) + 1)
            n_select (optional): An integer array indicating the number
                of references selected in each trial. Values must be
                greater than zero but less than the number of
                references for the corresponding trial.
                shape = n_trial,)
            is_ranked (optional): A Boolean array indicating which
                trials require reference selections to be ranked.
                shape = (n_trial,)
        """
        stimulus_set = self._check_stimulus_set(stimulus_set)

        self.n_trial = stimulus_set.shape[0]
        n_reference = self._infer_n_reference(stimulus_set)
        self.n_reference = self._check_n_reference(n_reference)

        # Format stimulus set.
        self.max_n_reference = np.amax(self.n_reference)
        self.stimulus_set = stimulus_set[:, 0:self.max_n_reference+1]

        if n_select is None:
            n_select = np.ones((self.n_trial), dtype=np.int32)
        else:
            n_select = self._check_n_select(n_select)
        self.n_select = n_select

        if is_ranked is None:
            is_ranked = np.full((self.n_trial), True)
        else:
            is_ranked = self._check_is_ranked(is_ranked)
        self.is_ranked = is_ranked

        # Attributes determined by concrete class.
        self.config_idx = None
        self.config_list = None
        self.outcome_idx_list = None

    def _check_stimulus_set(self, stimulus_set):
        """Check the argument `stimulus_set`.

        Raises:
            ValueError

        """
        if not issubclass(stimulus_set.dtype.type, np.integer):
            raise ValueError((
                "The argument `stimulus_set` must be a 2D array of "
                "integers."))
        if np.sum(np.less(stimulus_set, -1)) > 0:
            raise ValueError((
                "The argument `stimulus_set` must only contain integers "
                "greater than or equal to -1."))
        return stimulus_set

    def _infer_n_reference(self, stimulus_set):
        """Return the number of references in each trial.

        Infers the number of available references for each trial. The
        function assumes that values less than zero, are placeholder
        values and should be treated as non-existent.

        Arguments:
            stimulus_set: shape = [n_trial, 1]

        Returns:
            n_reference: An integer array indicating the number of
                references in each trial.
                shape = [n_trial, 1]

        """
        max_ref = stimulus_set.shape[1] - 1
        n_reference = max_ref - np.sum(stimulus_set < 0, axis=1)
        return n_reference.astype(dtype=np.int32)

    def _check_n_reference(self, n_reference):
        """Check the argument `n_reference`.

        Raises:
            ValueError

        """
        if np.sum(np.less(n_reference, 2)) > 0:
            raise ValueError((
                "The argument `stimulus_set` must contain at least three "
                "non-negative integers per a row, i.e. one query and at least "
                "two reference stimuli per trial."))
        return n_reference

    def _check_n_select(self, n_select):
        """Check the argument `n_select`.

        Raises:
            ValueError

        """
        n_select = n_select.astype(np.int32)
        # Check shape argreement.
        if not (n_select.shape[0] == self.n_trial):
            raise ValueError((
                "The argument `n_select` must have the same length as the "
                "number of rows in the argument 'stimulus_set'."))
        # Check lowerbound support limit.
        bad_locs = n_select < 1
        n_bad = np.sum(bad_locs)
        if n_bad != 0:
            raise ValueError((
                "The argument `n_select` contains integers less than 1. "
                "Found {0} bad trial(s).").format(n_bad))
        # Check upperbound support limit.
        bad_locs = np.greater_equal(n_select, self.n_reference)
        n_bad = np.sum(bad_locs)
        if n_bad != 0:
            raise ValueError((
                "The argument `n_select` contains integers greater than "
                "or equal to the corresponding 'n_reference'. Found {0} bad "
                "trial(s).").format(n_bad))
        return n_select

    def _check_is_ranked(self, is_ranked):
        """Check the argument `is_ranked`.

        Raises:
            ValueError

        """
        if not (is_ranked.shape[0] == self.n_trial):
            raise ValueError((
                "The argument `n_select` must have the same length as the "
                "number of rows in the argument 'stimulus_set'."))
        bad_locs = np.not_equal(is_ranked, True)
        n_bad = np.sum(bad_locs)
        if n_bad != 0:
            raise ValueError((
                "The unranked version is not implemented, Found {0} bad "
                "trial(s).").format(n_bad))
        return is_ranked

    @abstractmethod
    def _set_configuration_data(self, *args):
        """Generate a unique ID for each trial configuration.

        Helper function that generates a unique ID for each of the
        unique trial configurations in the provided data set.

        Notes:
            Sets three attributes of object.
            config_idx: A unique index for each type of trial
                configuration.
            config_list: A DataFrame containing all the unique
                trial configurations.
            outcome_idx_list: A list of the possible outcomes for each
                trial configuration.

        """
        pass

    @abstractmethod
    def subset(self, index):
        """Return subset of trials as new SimilarityTrials object.

        Arguments:
            index: The indices corresponding to the subset.

        Returns:
            A new SimilarityTrials object.

        """
        pass

    def outcome_tensor(self):
        """Return outcome tensor."""
        # TODO better doc, test
        n_config = self.config_list.shape[0]

        n_outcome_list = self.config_list['n_outcome'].values
        max_n_outcome = np.max(n_outcome_list)

        n_reference_list = self.config_list['n_reference'].values
        max_n_reference = np.max(n_reference_list)

        outcome_tensor = -1 * np.ones(
            (n_config, max_n_outcome, max_n_reference), dtype=np.int32)
        for i_config in range(n_config):
            outcome_tensor[
                i_config,
                0:n_outcome_list[i_config],
                0:n_reference_list[i_config]] = self.outcome_idx_list[i_config]
        return outcome_tensor

    @abstractmethod
    def save(self, filepath):
        """Save the SimilarityTrials object as an HDF5 file.

        Arguments:
            filepath: String specifying the path to save the data.

        """
        pass


class Docket(SimilarityTrials):
    """Object that encapsulates unjudged similarity trials.

    The attributes and behavior of Docket is largely inherited
    from SimilarityTrials.

    Attributes:
        n_trial: An integer indicating the number of trials.
        stimulus_set: An integer matrix containing indices that
            indicate the set of stimuli used in each trial. Each row
            indicates the stimuli used in one trial. The first column
            is the query stimulus. The remaining, columns indicate
            reference stimuli. Negative integers are used as
            placeholders to indicate non-existent references.
            shape = (n_trial, max(n_reference) + 1)
        n_reference: An integer array indicating the number of
            references in each trial.
            shape = (n_trial,)
        n_select: An integer array indicating the number of references
            selected in each trial.
            shape = (n_trial,)
        is_ranked: A Boolean array indicating which trials require
            reference selections to be ranked.
            shape = (n_trial,)
        config_idx: An integer array indicating the
            configuration of each trial. The integer is an index
            referencing the row of config_list and the element of
            outcome_idx_list.
            shape = (n_trial,)
        config_list: A DataFrame object describing the unique trial
            configurations. The columns are 'n_reference',
            'n_select', 'is_ranked'. and 'n_outcome'.
        outcome_idx_list: A list of 2D arrays indicating all possible
            outcomes for a trial configuration. Each element in the
            list corresponds to a trial configuration in config_list.
            Each row of the 2D array indicates one potential outcome.
            The values in the rows are the indices of the the reference
            stimuli (as specified in the attribute `stimulus_set`.

    Notes:
        stimulus_set: The order of the reference stimuli is
            unimportant. As usual, the the first column contains
            indices indicating query stimulus. The remaining columns
            contain indices indicating the reference stimuli in any
            order.
        Unique configurations and configuration IDs are determined by
            'n_reference', 'n_select', and 'is_ranked'.

    Methods:
        save: Save the Docket object to disk.
        subset: Return a subset of unjudged trials given an index.

    """

    def __init__(self, stimulus_set, n_select=None, is_ranked=None):
        """Initialize.

        Extends initialization of SimilarityTrials.

        Arguments:
            stimulus_set: The order of the reference indices is not
                important. See SimilarityTrials.
            n_select (optional): See SimilarityTrials.
            is_ranked (optional): See SimilarityTrials.
        """
        SimilarityTrials.__init__(self, stimulus_set, n_select, is_ranked)

        # Determine unique display configurations.
        self._set_configuration_data(
            self.n_reference, self.n_select, self.is_ranked)

    def subset(self, index):
        """Return subset of trials as a new Docket object.

        Arguments:
            index: The indices corresponding to the subset.

        Returns:
            A new Docket object.

        """
        return Docket(
            self.stimulus_set[index, :], self.n_select[index],
            self.is_ranked[index]
        )

    def _set_configuration_data(self, n_reference, n_select, is_ranked):
        """Generate a unique ID for each trial configuration.

        Helper function that generates a unique ID for each of the
        unique trial configurations in the provided data set.

        Arguments:
            n_reference: An integer array indicating the number of
                references in each trial.
                shape = (n_trial,)
            n_select: An integer array indicating the number of
                references selected in each trial.
                shape = (n_trial,)
            is_ranked:  Boolean array indicating which trials had
                selected references that were ordered.
                shape = (n_trial,)

        Notes:
            Sets three attributes of object.
            config_idx: A unique index for each type of trial
                configuration.
            config_list: A DataFrame containing all the unique
                trial configurations.
            outcome_idx_list: A list of the possible outcomes for each
                trial configuration.

        """
        n_trial = len(n_reference)

        # Determine unique display configurations.
        n_outcome_placeholder = np.zeros(n_select.shape[0], dtype=np.int32)
        d = {
            'n_reference': n_reference, 'n_select': n_select,
            'is_ranked': is_ranked, 'n_outcome': n_outcome_placeholder
            }
        df_config = pd.DataFrame(d)
        df_config = df_config.drop_duplicates()
        n_config = len(df_config)
        n_out_idx = df_config.columns.get_loc("n_outcome")

        # Assign display configuration ID for every observation.
        config_idx = np.empty(n_trial, dtype=np.int32)
        outcome_idx_list = []
        for i_config in range(n_config):
            # Determine number of possible outcomes for configuration.
            outcome_idx = _possible_outcomes(df_config.iloc[i_config])
            outcome_idx_list.append(outcome_idx)
            n_outcome = outcome_idx.shape[0]
            df_config.iloc[i_config, n_out_idx] = n_outcome
            # Find trials matching configuration.
            a = (n_reference == df_config['n_reference'].iloc[i_config])
            b = (n_select == df_config['n_select'].iloc[i_config])
            c = (is_ranked == df_config['is_ranked'].iloc[i_config])
            f = np.array((a, b, c))
            display_type_locs = np.all(f, axis=0)
            config_idx[display_type_locs] = i_config

        self.config_idx = config_idx
        self.config_list = df_config
        self.outcome_idx_list = outcome_idx_list

    def save(self, filepath):
        """Save the Docket object as an HDF5 file.

        Arguments:
            filepath: String specifying the path to save the data.

        """
        f = h5py.File(filepath, "w")
        f.create_dataset("trial_type", data="Docket")
        f.create_dataset("stimulus_set", data=self.stimulus_set)
        f.create_dataset("n_select", data=self.n_select)
        f.create_dataset("is_ranked", data=self.is_ranked)
        f.close()


class Observations(SimilarityTrials):
    """Object that encapsulates judged similarity trials.

    The attributes and behavior of Observations are largely inherited
    from SimilarityTrials.

    Attributes:
        n_trial: An integer indicating the number of trials.
        stimulus_set: An integer matrix containing indices that
            indicate the set of stimuli used in each trial. Each row
            indicates the stimuli used in one trial. The first column
            is the query stimulus. The remaining, columns indicate
            reference stimuli. Negative integers are used as
            placeholders to indicate non-existent references.
            shape = (n_trial, max(n_reference) + 1)
        n_reference: An integer array indicating the number of
            references in each trial.
            shape = (n_trial,)
        n_select: An integer array indicating the number of references
            selected in each trial.
            shape = (n_trial,)
        is_ranked: A Boolean array indicating which trials require
            reference selections to be ranked.
            shape = (n_trial,)
        config_idx: An integer array indicating the
            configuration of each trial. The integer is an index
            referencing the row of config_list and the element of
            outcome_idx_list.
            shape = (n_trial,)
        config_list: A DataFrame object describing the unique trial
            configurations.
        outcome_idx_list: A list of 2D arrays indicating all possible
            outcomes for a trial configuration. Each element in the
            list corresponds to a trial configuration in config_list.
            Each row of the 2D array indicates one potential outcome.
            The values in the rows are the indices of the the reference
            stimuli (as specified in the attribute `stimulus_set`.
        group_id: An integer array indicating the group membership of
            each trial. It is assumed that group_id is composed of
            integers from [0, M-1] where M is the total number of
            groups.
            shape = (n_trial,)
        agent_id: An integer array indicating the agent ID of a trial.
            It is assumed that all IDs are non-negative and that
            observations with the same agent ID were judged by a single
            agent.
            shape = (n_trial,)
        session_id: An integer array indicating the session ID of a
            trial. It is assumed that all IDs are non-negative. Trials
            with different session IDs were obtained during different
            sessions.
            shape = (n_trial,)
        weight: An float array indicating the inference weight of each
            trial.
            shape = (n_trial,)
        rt_ms: An array indicating the response time (in milliseconds)
            of the agent for each trial.
        session_id: An integer array indicating the session ID of
            a trial. It is assumed that observations with the same
            session ID were judged by a single agent. A single agent
            may have completed multiple sessions.
            shape = (n_trial,) TODO MAYBE

    Notes:
        stimulus_set: The order of the reference stimuli is important.
            As usual, the the first column contains indices indicating
            query stimulus. The remaining columns contain indices
            indicating the reference stimuli. An agent's selected
            references are listed first (in order of selection if the
            trial is ranked) and remaining unselected references are
            listed in any order.
        Unique configurations and configuration IDs are determined by
            'group_id' in addition to the usual 'n_reference',
            'n_select', and 'is_ranked' variables.

    Methods:
        subset: Return a subset of judged trials given an index.
        set_group_id: Override the group ID of all trials.
        set_weight: Override the weight of all trials.
        save: Save the observations data structure to disk.

    """

    def __init__(self, stimulus_set, n_select=None, is_ranked=None,
                 group_id=None, agent_id=None, session_id=None, weight=None,
                 rt_ms=None):
        """Initialize.

        Extends initialization of SimilarityTrials.

        Arguments:
            stimulus_set: The order of reference indices is important.
                An agent's selected references are listed first (in
                order of selection if the trial is ranked) and
                remaining unselected references are listed in any
                order. See SimilarityTrials.
            n_select (optional): See SimilarityTrials.
            is_ranked (optional): See SimilarityTrials.
            group_id (optional): An integer array indicating the group
                membership of each trial. It is assumed that group_id
                is composed of integers from [0, M-1] where M is the
                total number of groups.
                shape = (n_trial,)
            agent_id: An integer array indicating the agent ID of a
                trial. It is assumed that all IDs are non-negative and
                that observations with the same agent ID were judged by
                a single agent.
                shape = (n_trial,)
            session_id: An integer array indicating the session ID of a
                trial. It is assumed that all IDs are non-negative.
                Trials with different session IDs were obtained during
                different sessions.
            shape = (n_trial,)
            weight (optional): A float array indicating the inference
                weight of each trial.
                shape = (n_trial,1)
            rt_ms(optional): An array indicating the response time
                (in milliseconds) of the agent for each trial.
                shape = (n_trial,1)

        """
        SimilarityTrials.__init__(self, stimulus_set, n_select, is_ranked)

        # Handle default settings.
        if group_id is None:
            group_id = np.zeros((self.n_trial), dtype=np.int32)
        else:
            group_id = self._check_group_id(group_id)
        self.group_id = group_id

        if agent_id is None:
            agent_id = np.zeros((self.n_trial), dtype=np.int32)
        else:
            agent_id = self._check_agent_id(agent_id)
        self.agent_id = agent_id

        if session_id is None:
            session_id = np.zeros((self.n_trial), dtype=np.int32)
        else:
            session_id = self._check_session_id(session_id)
        self.session_id = session_id

        if weight is None:
            weight = np.ones((self.n_trial))
        else:
            weight = self._check_weight(weight)
        self.weight = weight

        if rt_ms is None:
            rt_ms = -np.ones((self.n_trial))
        else:
            rt_ms = self._check_rt(rt_ms)
        self.rt_ms = rt_ms

        # Determine unique display configurations.
        self._set_configuration_data(
            self.n_reference, self.n_select, self.is_ranked, group_id)

    def _check_group_id(self, group_id):
        """Check the argument group_id."""
        group_id = group_id.astype(np.int32)
        # Check shape argreement.
        if not (group_id.shape[0] == self.n_trial):
            raise ValueError((
                "The argument 'group_id' must have the same length as the "
                "number of rows in the argument 'stimulus_set'."))
        # Check lowerbound support limit.
        bad_locs = group_id < 0
        n_bad = np.sum(bad_locs)
        if n_bad != 0:
            raise ValueError((
                "The parameter 'group_id' contains integers less than 0. "
                "Found {0} bad trial(s).").format(n_bad))
        return group_id

    def _check_agent_id(self, agent_id):
        """Check the argument agent_id."""
        agent_id = agent_id.astype(np.int32)
        # Check shape argreement.
        if not (agent_id.shape[0] == self.n_trial):
            raise ValueError((
                "The argument 'agent_id' must have the same length as the "
                "number of rows in the argument 'stimulus_set'."))
        # Check lowerbound support limit.
        bad_locs = agent_id < 0
        n_bad = np.sum(bad_locs)
        if n_bad != 0:
            raise ValueError((
                "The parameter 'agent_id' contains integers less than 0. "
                "Found {0} bad trial(s).").format(n_bad))
        return agent_id

    def _check_session_id(self, session_id):
        """Check the argument session_id."""
        session_id = session_id.astype(np.int32)
        # Check shape argreement.
        if not (session_id.shape[0] == self.n_trial):
            raise ValueError((
                "The argument 'session_id' must have the same length as the "
                "number of rows in the argument 'stimulus_set'."))
        # Check lowerbound support limit.
        bad_locs = session_id < 0
        n_bad = np.sum(bad_locs)
        if n_bad != 0:
            raise ValueError((
                "The parameter 'session_id' contains integers less than 0. "
                "Found {0} bad trial(s).").format(n_bad))
        return session_id

    def _check_weight(self, weight):
        """Check the argument weight."""
        weight = weight.astype(np.float)
        # Check shape argreement.
        if not (weight.shape[0] == self.n_trial):
            raise ValueError((
                "The argument 'weight' must have the same length as the "
                "number of rows in the argument 'stimulus_set'."))
        return weight

    def _check_rt(self, rt_ms):
        """Check the argument rt_ms."""
        rt_ms = rt_ms.astype(np.float)
        # Check shape argreement.
        if not (rt_ms.shape[0] == self.n_trial):
            raise ValueError((
                "The argument 'rt_ms' must have the same length as the "
                "number of rows in the argument 'stimulus_set'."))
        return rt_ms

    def subset(self, index):
        """Return subset of trials as a new Observations object.

        Arguments:
            index: The indices corresponding to the subset.

        Returns:
            A new Observations object.

        """
        return Observations(
            self.stimulus_set[index, :], n_select=self.n_select[index],
            is_ranked=self.is_ranked[index], group_id=self.group_id[index],
            agent_id=self.agent_id[index], session_id=self.session_id[index],
            weight=self.weight[index], rt_ms=self.rt_ms[index]
        )

    def _set_configuration_data(
                self, n_reference, n_select, is_ranked, group_id,
                session_id=None):
        """Generate a unique ID for each trial configuration.

        Helper function that generates a unique ID for each of the
        unique trial configurations in the provided data set.

        Arguments:
            n_reference: An integer array indicating the number of
                references in each trial.
                shape = (n_trial,)
            n_select: An integer array indicating the number of
                references selected in each trial.
                shape = (n_trial,)
            is_ranked:  Boolean array indicating which trials had
                selected references that were ordered.
                shape = (n_trial,)
            group_id: An integer array indicating the group membership
                of each trial. It is assumed that group is composed of
                integers from [0, M-1] where M is the total number of
                groups. Separate attention weights are inferred for
                each group.
                shape = (n_trial,)
            session_id: An integer array indicating the session ID of
                a trial. It is assumed that observations with the same
                session ID were judged by a single agent. A single
                agent may have completed multiple sessions.
                shape = (n_trial,)

        Notes:
            Sets three attributes of object.
            config_idx: A unique index for each type of trial
                configuration.
            config_list: A DataFrame containing all the unique
                trial configurations.
            outcome_idx_list: A list of the possible outcomes for each
                trial configuration.

        """
        n_trial = len(n_reference)

        if session_id is None:
            session_id = np.zeros((n_trial), dtype=np.int32)

        # Determine unique display configurations.
        n_outcome_placeholder = np.zeros(n_select.shape[0], dtype=np.int32)
        d = {
            'n_reference': n_reference, 'n_select': n_select,
            'is_ranked': is_ranked, 'group_id': group_id,
            'session_id': session_id, 'n_outcome': n_outcome_placeholder
            }
        df_config = pd.DataFrame(d)
        df_config = df_config.drop_duplicates()
        n_config = len(df_config)
        n_out_idx = df_config.columns.get_loc("n_outcome")

        # Assign display configuration index for every observation.
        config_idx = np.empty(n_trial, dtype=np.int32)
        outcome_idx_list = []
        for i_config in range(n_config):
            # Determine number of possible outcomes for configuration.
            outcome_idx = _possible_outcomes(df_config.iloc[i_config])
            outcome_idx_list.append(outcome_idx)
            n_outcome = outcome_idx.shape[0]
            df_config.iloc[i_config, n_out_idx] = n_outcome
            # Find trials matching configuration.
            a = (n_reference == df_config['n_reference'].iloc[i_config])
            b = (n_select == df_config['n_select'].iloc[i_config])
            c = (is_ranked == df_config['is_ranked'].iloc[i_config])
            d = (group_id == df_config['group_id'].iloc[i_config])
            e = (session_id == df_config['session_id'].iloc[i_config])
            f = np.array((a, b, c, d, e))
            display_type_locs = np.all(f, axis=0)
            config_idx[display_type_locs] = i_config

        self.config_idx = config_idx
        self.config_list = df_config
        self.outcome_idx_list = outcome_idx_list

    def set_group_id(self, group_id):
        """Override the existing group_ids.

        Arguments:
            group_id: The new group IDs. Can be an integer or an array
                of integers with shape=(self.n_trial,).

        """
        if np.isscalar(group_id):
            group_id = group_id * np.ones((self.n_trial), dtype=np.int32)
        else:
            group_id = self._check_group_id(group_id)
        self.group_id = copy.copy(group_id)

        # Re-derive unique display configurations.
        self._set_configuration_data(
            self.n_reference, self.n_select, self.is_ranked, group_id)

    def set_weight(self, weight):
        """Override the existing group_ids.

        Arguments:
            weight: The new weight. Can be an float or an array
                of floats with shape=(self.n_trial,).
        """
        if np.isscalar(weight):
            weight = weight * np.ones((self.n_trial), dtype=np.int32)
        else:
            weight = self._check_weight(weight)
        self.weight = copy.copy(weight)

    def save(self, filepath):
        """Save the Docket object as an HDF5 file.

        Arguments:
            filepath: String specifying the path to save the data.

        """
        f = h5py.File(filepath, "w")
        f.create_dataset("trial_type", data="Observations")
        f.create_dataset("stimulus_set", data=self.stimulus_set)
        f.create_dataset("n_select", data=self.n_select)
        f.create_dataset("is_ranked", data=self.is_ranked)
        f.create_dataset("group_id", data=self.group_id)
        f.create_dataset("agent_id", data=self.agent_id)
        f.create_dataset("session_id", data=self.session_id)
        f.create_dataset("weight", data=self.weight)
        f.create_dataset("rt_ms", data=self.rt_ms)
        f.close()


def stack(trials_list):
        """Return a SimilarityTrials object containing all trials.

        The stimulus_set of each SimilarityTrials object is padded
        first to match the maximum number of references of all the
        objects.

        Arguments:
            trials_list: A tuple of SimilarityTrials objects to be
                stacked.

        Returns:
            A new SimilarityTrials object.

        """
        # Determine the maximum number of references.
        max_n_reference = 0
        for i_trials in trials_list:
            if i_trials.max_n_reference > max_n_reference:
                max_n_reference = i_trials.max_n_reference

        # Grab relevant information from first entry in list.
        stimulus_set = _pad_stimulus_set(
            trials_list[0].stimulus_set,
            max_n_reference
        )
        n_select = trials_list[0].n_select
        is_ranked = trials_list[0].is_ranked
        is_judged = True
        try:
            group_id = trials_list[0].group_id
            agent_id = trials_list[0].agent_id
            session_id = trials_list[0].session_id
            weight = trials_list[0].weight
            rt_ms = trials_list[0].rt_ms
        except AttributeError:
            is_judged = False

        for i_trials in trials_list[1:]:
            stimulus_set = np.vstack((
                stimulus_set,
                _pad_stimulus_set(i_trials.stimulus_set, max_n_reference)
            ))
            n_select = np.hstack((n_select, i_trials.n_select))
            is_ranked = np.hstack((is_ranked, i_trials.is_ranked))
            if is_judged:
                group_id = np.hstack((group_id, i_trials.group_id))
                agent_id = np.hstack((agent_id, i_trials.agent_id))
                session_id = np.hstack((session_id, i_trials.session_id))
                weight = np.hstack((weight, i_trials.weight))
                rt_ms = np.hstack((rt_ms, i_trials.rt_ms))

        if is_judged:
            trials_stacked = Observations(
                stimulus_set, n_select=n_select, is_ranked=is_ranked,
                group_id=group_id, agent_id=agent_id, session_id=session_id,
                weight=weight, rt_ms=rt_ms
            )
        else:
            trials_stacked = Docket(
                stimulus_set, n_select, is_ranked
            )
        return trials_stacked


def squeeze(sim_trials, mode="sg"):
    """Squeeze indices in trials to be small and consecutive.

    Indices are reset to be between 0 and N-1 where N is the number of
    unique stimuli used in sim_trials.

    Arguments:
        sim_trials: A SimilarityTrials object.
        mode (optional): The mode in which to squueze the indices.

    Returns:
        sim_trials_sq: A SimilarityTrials object.

    """
    unique_stimuli = np.unique(sim_trials.stimulus_set)

    # Remove placeholder value from list of unique stimuli indices.
    loc = np.equal(unique_stimuli, -1)
    unique_stimuli = unique_stimuli[np.logical_not(loc)]

    # Squeeze trials.
    sim_trials_sq = copy.deepcopy(sim_trials)
    for new_idx, old_idx in enumerate(unique_stimuli):
        locs = np.equal(sim_trials.stimulus_set, old_idx)
        sim_trials_sq.stimulus_set[locs] = new_idx

    # MAYBE Squeeze groups.
    return sim_trials_sq, unique_stimuli


def load(filepath, verbose=0):
    """Load data saved via the save method.

    The loaded data is instantiated as a concrete class of
    SimilarityTrials.

    Arguments:
        filepath: The location of the hdf5 file to load.
        verbose (optional): Controls the verbosity of printed summary.

    Returns:
        Loaded trials.

    Raises:
        ValueError

    """
    f = h5py.File(filepath, "r")
    # Common attributes.
    trial_type = f["trial_type"][()]
    stimulus_set = f["stimulus_set"][()]
    n_select = f["n_select"][()]
    is_ranked = f["is_ranked"][()]

    if trial_type == "Docket":
        loaded_trials = Docket(
            stimulus_set, n_select=n_select, is_ranked=is_ranked
        )
    elif trial_type == "Observations":
        # Observations specific attributes.
        group_id = f["group_id"][()]

        # For backwards compatability.
        if "weight" in f:
            weight = f["weight"][()]
        else:
            weight = np.ones((len(n_select)))
        if "rt_ms" in f:
            rt_ms = f["rt_ms"][()]
        else:
            rt_ms = -np.ones((len(n_select)))
        if "agent_id" in f:
            agent_id = f["agent_id"][()]
        else:
            agent_id = np.zeros((len(n_select)))
        if "session_id" in f:
            session_id = f["session_id"][()]
        else:
            session_id = np.zeros((len(n_select)))
        loaded_trials = Observations(
            stimulus_set, n_select=n_select, is_ranked=is_ranked,
            group_id=group_id, agent_id=agent_id, session_id=session_id,
            weight=weight, rt_ms=rt_ms
        )
    else:
        raise ValueError('No class found matching the provided `trial_type`.')
    f.close()

    if verbose > 0:
        print("Trial Summary")
        print('  trial_type: {0}'.format(trial_type))
        print('  n_trial: {0}'.format(loaded_trials.n_trial))
        if trial_type == "Observations":
            print(
                '  n_agent: {0}'.format(
                    len(np.unique(loaded_trials.agent_id))
                )
            )
            print(
                '  n_group: {0}'.format(
                    len(np.unique(loaded_trials.group_id))
                )
            )
        print('')
    return loaded_trials


load_trials = load


def _pad_stimulus_set(stimulus_set, max_n_reference):
    """Pad 2D array with columns composed of -1."""
    n_trial = stimulus_set.shape[0]
    n_pad = max_n_reference - (stimulus_set.shape[1] - 1)
    if n_pad > 0:
        pad_mat = -1 * np.ones((n_trial, n_pad), dtype=np.int32)
        stimulus_set = np.hstack((stimulus_set, pad_mat))
    return stimulus_set


def _possible_outcomes(trial_configuration):
    """Return the possible outcomes of a trial configuration.

    Arguments:
        trial_configuration: A trial configuration Pandas Series.

    Returns:
        An 2D array indicating all possible outcomes where the values
            indicate indices of the reference stimuli. Each row
            corresponds to one outcome. Note the indices refer to
            references only and does not include an index for the
            query. Also note that the unpermuted index is returned
            first.

    """
    n_reference = int(trial_configuration['n_reference'])
    n_select = int(trial_configuration['n_select'])

    reference_list = range(n_reference)

    # Get all permutations of length n_select.
    perm = permutations(reference_list, n_select)

    selection = list(perm)
    n_outcome = len(selection)

    outcomes = np.empty((n_outcome, n_reference), dtype=np.int32)
    for i_outcome in range(n_outcome):
        # Fill in selections.
        outcomes[i_outcome, 0:n_select] = selection[i_outcome]
        # Fill in unselected.
        dummy_idx = np.arange(n_reference)
        for i_selected in range(n_select):
            loc = dummy_idx != outcomes[i_outcome, i_selected]
            dummy_idx = dummy_idx[loc]

        outcomes[i_outcome, n_select:] = dummy_idx

    return outcomes



"""Package initialization file.

Modules:
    benchmark
    catalog
    datasets
    dimensionality
    generator
    models
    preprocess
    simulate
    trials
    utils
    visualize
"""

import psiz.catalog
import psiz.datasets
import psiz.dimensionality
import psiz.generator
import psiz.models
import psiz.preprocess
import psiz.simulate
import psiz.trials
import psiz.utils
import psiz.visualize
